# flake8: noqa

"""The Simple Management Protocol (SMP) for remotely managing MCU firmware.

This package implements de/serialization of SMP messages allowing for their use
on the transport of your choice.

The SMP specification can be found [here](https://docs.zephyrproject.org/latest/services/device_mgmt/smp_protocol.html).

## Usage

SMP messages are represented as Pydantic models.  Each SMP Request and
Response contains a `header` attribute that is an instance of
`smp.header.Header`.  Other attributes are specific to the message type.

For example, to create a `smp.os_management.EchoWriteRequest`:

```python
from smp.os_management import EchoWriteRequest

request = EchoWriteRequest(d="Hello world!")
print(bytes(request))
```

Prints the serialized SMP Frame:
```python
b'\\n\\x00\\x00\\x10\\x00\\x00\\x00\\x00\\xa1adlHello world!'
```

All messages can be deserialized and validated using the `loads()` method.  To 
load a `smp.os_management.EchoWriteResponse`:

```python
from smp.os_management import EchoWriteResponse

data = bytes.fromhex("0b00000700008c00a1617263486921")  # data from the transport
response = EchoWriteResponse.loads(data)
print(response)
```

Prints the deserialized SMP message representation:
```
header=Header(op=<OP.WRITE_RSP: 3>, version=<Version.V2: 1>, flags=<Flag: 0>,
length=7, group_id=0, sequence=140, command_id=0) version=<Version.V2: 1>
sequence=140 smp_data=b'\\x0b\\x00\\x00\\x07\\x00\\x00\\x8c\\x00\\xa1arcHi!' r='Hi!'
```
Generally, the `header` can be ignored and the message-specific attributes are
what you are interested in.
```
# print(response.r)
Hi!
```
All models and their attributes are statically typed and validated; enforced by
mypy linting and by Pydantic at runtime.

## Serialization

There are a few optional arguments that are common to all SMP messages when they
are created.

-   `header` is the SMP header.  Typically this can be left as `None` and will be
    so that the header is created automatically.
-   `version` is the SMP version.  This defaults to `smp.header.Version.V2`.
-   `sequence` is the sequence number of the message.  If not provided, it will
    be automatically generated using an incrementing counter.

Take a look at `smp.message` for more information on the base classes.

## Deserialization

If you are writing an SMP client, then you already know the type of the
incoming message because it must be a Response to your Request, or an
`smp.error.ErrorV1` or `smp.error.ErrorV2`.  You can use the
`smp.message._MessageBase.loads()` method that is common to all SMP
messages to deserialize and validate the message.

If you are writing an SMP server, then Python and SMP are odd choices!  Yet, you
can narrow the type by first loading the header with `smp.header.Header.loads()`.

## Encoding & Decoding

The USB and serial transports defined by Zephyr use a base64 encoding and
framing.  The encoding/fragmentation and decoding/reassembly is provided by 
`smp.packet.encode()` and `smp.packet.decode()`.

More information is at the [Zephyr docs](https://docs.zephyrproject.org/latest/services/device_mgmt/smp_transport.html#uart-serial-and-console)

## Typing

This package is meticulously typed and is intended to be used with mypy.

## Validation

All models are validated in order to detect transport and SMP server errors.  It
is impossible to create an invalid SMP message or deserialize an invalid SMP
message.  If you find a way, please open an issue.
"""



"""The Simple Management Protocol (SMP) OS Management group."""

from __future__ import annotations

from enum import IntEnum, unique
from typing import Any, Dict, Literal

from pydantic import BaseModel, ConfigDict, Field

from smp import error, header, message


class EchoWriteRequest(message.WriteRequest):
    """Echo back the provided string."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.ECHO

    d: str
    """String to echo."""


class EchoWriteResponse(message.WriteResponse):
    """Success response to an echo request."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.ECHO

    r: str
    """Echoed string."""


class ResetWriteRequest(message.WriteRequest):
    """Performs reset of system.

    The device should issue response before resetting so that the SMP client
    could receive information that the command has been accepted. By default,
    this command is accepted in all conditions, however if the
    `CONFIG_MCUMGR_GRP_OS_RESET_HOOK` is enabled and an application registers a
    callback, the callback will be called when this command is issued and can be
    used to perform any necessary tidy operations prior to the module rebooting,
    or to reject the reset request outright altogether with an error response.

    For details on this functionality, see [callbacks](https://docs.zephyrproject.org/latest/services/device_mgmt/mcumgr_callbacks.html#mcumgr-callbacks).  # noqa: E501
    """

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.RESET

    force: Literal[0, 1] | None = None
    """Force reset.

    Normally the command sends an empty CBOR map as data, but if a previous
    reset attempt has responded with “rc” equal to MGMT_ERR_EBUSY then the
    following map may be sent to force a reset
    """


class ResetWriteResponse(message.WriteResponse):
    """Success response to a reset request."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.RESET


class TaskStatisticsReadRequest(message.ReadRequest):
    """Request task statistics."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.TASK_STATS


class TaskStatistics(BaseModel):
    """Task statistics."""

    model_config = ConfigDict(extra="forbid", frozen=True)

    prio: int
    """Task priority."""
    tid: int
    """Numeric task ID."""
    state: int
    """Numeric task state."""
    stkuse: int
    """Stack usage.

    The unit is system dependent and in case of Zephyr this is number of 4 byte words.
    """
    stksize: int
    """Stack size.

    The unit is system dependent and in case of Zephyr this is number of 4 byte words.
    """
    cswcnt: int
    """Number of context switches."""
    runtime: int
    """Runtime in ticks."""
    last_checkin: int
    """Set to 0 by Zephyr."""
    next_checkin: int
    """Set to 0 by Zephyr."""


class TaskStatisticsReadResponse(message.ReadResponse):
    """Task statistics response."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.TASK_STATS

    tasks: Dict[str, TaskStatistics]
    """Task statistics map."""


class MemoryPoolStatisticsReadRequest(message.ReadRequest):
    """Request memory pool statistics."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.MEMORY_POOL_STATS


class MemoryPoolStatistics(BaseModel):
    """Memory pool statistics."""

    model_config = ConfigDict(extra="forbid", frozen=True)

    blksize: int
    """Size of the memory block in the pool."""
    nblks: int
    """Number of memory blocks in the pool."""
    nfree: int
    """Number of free memory blocks in the pool."""
    min: int
    """Lowest number of free blocks the pool reached during run-time."""


class MemoryPoolStatisticsReadResponse(message.ReadResponse):
    """The memory pools are accessed by name."""

    model_config = ConfigDict(extra="allow", frozen=True)

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.MEMORY_POOL_STATS


class DateTimeReadRequest(message.ReadRequest):
    """Request the current date and time."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.DATETIME_STRING


class DateTimeReadResponse(message.ReadResponse):
    """Response to a date and time request."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.DATETIME_STRING

    datetime: str


class DateTimeWriteRequest(message.WriteRequest):
    """Set the current date and time."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.DATETIME_STRING

    datetime: str


class DateTimeWriteResponse(message.WriteResponse):
    """Success response to a date and time request."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.DATETIME_STRING


class MCUMgrParametersReadRequest(message.ReadRequest):
    """Request MCU Manager parameters."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.MCUMGR_PARAMETERS


class MCUMgrParametersReadResponse(message.ReadResponse):
    """Success response to a MCU Manager parameters request."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.MCUMGR_PARAMETERS

    buf_size: int
    """Single SMP buffer size, this includes SMP header and CBOR payload."""
    buf_count: int
    """Number of SMP buffers."""


class OSApplicationInfoReadRequest(message.ReadRequest):
    """Request information about the application running on the device."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.OS_APPLICATION_INFO

    format: str | None = None
    """Format specifier of returned response.

    Fields are appended in their natural ascending index order, not the order
    of characters that are received by the command.

    Format specifiers:
    * `s` Kernel name
    * `n` Node name
    * `r` Kernel release
    * `v` Kernel version
    * `b` Build date and time (requires `CONFIG_MCUMGR_GRP_OS_INFO_BUILD_DATE_TIME`)
    * `m` Machine
    * `p` Processor
    * `i` Hardware platform
    * `o` Operating system
    * `a` All fields (shorthand for all above options)

    If this option is not provided, the `s` Kernel name option will be used.
    """


class OSApplicationInfoReadResponse(message.ReadResponse):
    """Success response to an application information request."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.OS_APPLICATION_INFO

    output: str
    """Text response including requested parameters."""


class BootloaderInformationReadRequest(message.ReadRequest):
    """Request bootloader information."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.BOOTLOADER_INFO

    query: str | None = None
    """Is string representing query for parameters.

    With no restrictions how the query looks like as processing of query is left
    for bootloader backend. If there is no query, then response will return string
    identifying the bootloader.

    MCUboot supports the query string,"mode".  The response to mode is of type
    `MCUbootMode`.
    """


@unique
class MCUbootMode(IntEnum):
    UNKNOWN = -1
    APPLICATION = 0
    SWAP_USING_SCRATCH = 1
    OVERWRITE_ONLY = 2
    SWAP_WITHOUT_SCRATCH = 3
    DIRECT_XIP_WITHOUT_REVERT = 4
    DIRECT_XIP_WITH_REVERT = 5
    RAM_LOADER = 6


class MCUbootModeQueryResponse(BaseModel):
    """Response to a MCUboot mode query."""

    model_config = ConfigDict(extra="forbid", frozen=True)

    mode: MCUbootMode
    no_downgrade: bool | None = Field(alias="no-downgrade", default=None)


class BootloaderInformationReadResponse(message.ReadResponse):
    """Success response to a bootloader information request."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT
    _COMMAND_ID = header.CommandId.OSManagement.BOOTLOADER_INFO

    bootloader: str
    """String identifying the bootloader.  For MCUboot it will be "MCUboot"."""
    response: MCUbootModeQueryResponse | Any | None = None
    """Response to “query”.

    This is optional and may be left out in case when query yields no response,
    SMP version 2 error code of `OS_MGMT_ERR_QUERY_YIELDS_NO_ANSWER` is
    expected. Response may have more than one parameter reported back or it may
    be a map, that is dependent on bootloader backend and query."""


@unique
class OS_MGMT_RET_RC(IntEnum):
    """OS Management return codes."""

    OK = 0
    """No error, this is implied if there is no ret value in the response."""

    UNKNOWN = 1
    """Unknown error occurred."""

    INVALID_FORMAT = 2
    """The provided format value is not valid."""

    QUERY_YIELDS_NO_ANSWER = 3
    """Query was not recognized."""

    RTC_NOT_SET = 4
    """RTC is not set."""

    RTC_COMMAND_FAILED = 5
    """RTC command failed."""


class OSManagementErrorV1(error.ErrorV1):
    """OS Management error response."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT


class OSManagementErrorV2(error.ErrorV2[OS_MGMT_RET_RC]):
    """OS Management error response."""

    _GROUP_ID = header.GroupId.OS_MANAGEMENT



"""The Simple Management Protocol (SMP) Zephyr Management group."""


from enum import IntEnum, unique

import smp.error as smperr
import smp.header as smphdr
import smp.message as smpmsg


class EraseStorageRequest(smpmsg.WriteRequest):
    """Erase the storage area."""

    _GROUP_ID = smphdr.GroupId.ZEPHYR_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.ZephyrManagement.ERASE_STORAGE


class EraseStorageResponse(smpmsg.WriteResponse):
    """Success response to a storage area erase."""

    _GROUP_ID = smphdr.GroupId.ZEPHYR_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.ZephyrManagement.ERASE_STORAGE


@unique
class ZEPHYRBASIC_MGMT_ERR(IntEnum):
    """Return codes for the Zephyr Management group."""

    OK = 0
    """No error, this is implied if there is no ret value in the response"""

    UNKNOWN = 1
    """Unknown error occurred."""

    FLASH_OPEN_FAILED = 2
    """Opening of the flash area has failed."""

    FLASH_CONFIG_QUERY_FAIL = 3
    """Querying the flash area parameters has failed."""

    FLASH_ERASE_FAILED = 4
    """Erasing the flash area has failed."""


class ZephyrManagementErrorV1(smperr.ErrorV1):
    """Error response to a Zephyr Management command."""

    _GROUP_ID = smphdr.GroupId.ZEPHYR_MANAGEMENT


class ZephyrManagementErrorV2(smperr.ErrorV2[ZEPHYRBASIC_MGMT_ERR]):
    """Error response to a Zephyr Management command."""

    _GROUP_ID = smphdr.GroupId.ZEPHYR_MANAGEMENT



"""The Simple Management Protocol (SMP) Enumeration Management group."""

from __future__ import annotations

from enum import IntEnum, unique
from typing import Tuple

from pydantic import BaseModel, ConfigDict

import smp.error as smperr
import smp.header as smphdr
import smp.message as smpmsg


class GroupCountRequest(smpmsg.ReadRequest):
    """Read the number of SMP server groups.

    Count of supported groups returns the total number of SMP command groups
    that a device supports.
    """

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.EnumManagement.GROUP_COUNT


class GroupCountResponse(smpmsg.ReadResponse):
    """SMP group count response."""

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.EnumManagement.GROUP_COUNT

    count: int
    """Contains the total number of supported SMP groups on the device."""


class ListOfGroupsRequest(smpmsg.ReadRequest):
    """List the available SMP groups."""

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.EnumManagement.LIST_OF_GROUPS


class ListOfGroupsResponse(smpmsg.ReadResponse):
    """SMP group list response."""

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.EnumManagement.LIST_OF_GROUPS

    groups: Tuple[smphdr.GroupIdField, ...]
    """Contains a list of the supported SMP group IDs on the device."""


class GroupIdRequest(smpmsg.ReadRequest):
    """List a SMP group by index.

    Fetch single group ID command allows listing the group IDs of supported SMP
    groups on the device, one by one.
    """

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.EnumManagement.GROUP_ID

    index: int | None = None
    """Contains the (0-based) index of the group to return information on, can
    be omitted to return the first group's details.
"""


class GroupIdResponse(smpmsg.ReadResponse):
    """SMP group at index response."""

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.EnumManagement.GROUP_ID

    group: smphdr.GroupIdField
    """The Group ID at the requested index."""
    end: bool | None = None
    """Will be set to true if the listed group is the final supported group on
    the device, otherwise will be omitted.
    """


class GroupDetailsRequest(smpmsg.ReadRequest):
    """Request the details of the supported SMP groups.

    Details on supported groups command allows fetching details on each
    supported SMP group, such as the name and number of handlers. A device can
    specify an allow list of groups to return details on or details on all
    groups can be returned.

    This command is optional, it can be enabled using
    `CONFIG_MCUMGR_GRP_ENUM_DETAILS`. The optional name and number of handlers
    can be enabled/disabled with `CONFIG_MCUMGR_GRP_ENUM_DETAILS_NAME` and
    `CONFIG_MCUMGR_GRP_ENUM_DETAILS_HANDLERS`.
    """

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.EnumManagement.GROUP_DETAILS

    groups: Tuple[smphdr.GroupIdField, ...] | None = None
    """Contains a list of the SMP group IDs to fetch details on.

    If omitted, details on all supported groups will be returned.
    """


class GroupDetails(BaseModel):
    """Group Details"""

    model_config = ConfigDict(extra="forbid", frozen=True)

    group: smphdr.GroupIdField
    """The group ID of the SMP command group."""
    name: str | None = None
    """The name of the SMP command group."""
    handlers: int | None = None
    """The number of handlers that the SMP command group supports."""


class GroupDetailsResponse(smpmsg.ReadResponse):
    """SMP group details response."""

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT
    _COMMAND_ID = smphdr.CommandId.EnumManagement.GROUP_DETAILS

    groups: Tuple[GroupDetails, ...]
    """Contains a list of the requested SMP group details."""


@unique
class ENUM_MGMT_ERR(IntEnum):
    """Return codes for the enumeration management group."""

    OK = 0
    """No error, this is implied if there is no ret value in the response."""

    UNKNOWN = 1
    """Unknown error occurred."""

    ERR_TOO_MANY_GROUP_ENTRIES = 2
    """Too many entries were provided."""

    ERR_INSUFFICIENT_HEAP_FOR_ENTRIES = 3
    """Insufficient heap memory to store entry data."""

    ENUM_MGMT_ERR_INDEX_TOO_LARGE = 4
    """Provided index is larger than the number of supported groups."""


class EnumManagementErrorV1(smperr.ErrorV1):
    """Error response to a enumeration management command."""

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT


class EnumManagementErrorV2(smperr.ErrorV2[ENUM_MGMT_ERR]):
    """Error response to a enumeration management command."""

    _GROUP_ID = smphdr.GroupId.ENUM_MANAGEMENT



"""The Simple Management Protocol (SMP) Intercreate Management group."""

from __future__ import annotations

from enum import IntEnum, unique

from smp import error, header, message


class ImageUploadWriteRequest(message.WriteRequest):
    """Upload an image to an application-defined location like a secondary MCU."""

    _GROUP_ID = header.UserGroupId.INTERCREATE
    _COMMAND_ID = header.CommandId.Intercreate.UPLOAD

    off: int
    """The offset in the image to write to."""

    data: bytes
    """The data to write to the image."""

    image: int | None = None
    """The image to write to; required when off == 0."""

    len: int | None = None
    """The length of the data to write; required when off == 0."""

    sha: bytes | None = None
    """The SHA-256 hash of the image; optional when off == 0, else ignored."""


class ImageUploadWriteResponse(message.WriteResponse):
    """Success response to an image upload request."""

    _GROUP_ID = header.UserGroupId.INTERCREATE
    _COMMAND_ID = header.CommandId.Intercreate.UPLOAD

    off: int
    """The offset in the image after the request was written."""


@unique
class IC_MGMT_ERR(IntEnum):
    """Intercreate Management error codes."""

    OK = 0
    """No error."""

    INVALID_IMAGE = 1
    """No image matched the image provided."""


class ErrorV1(error.ErrorV1):
    """Intercreate Management error response."""

    _GROUP_ID = header.UserGroupId.INTERCREATE


class ErrorV2(error.ErrorV2[IC_MGMT_ERR]):
    """Intercreate Management error response."""

    _GROUP_ID = header.UserGroupId.INTERCREATE



"""The Simple Management Protocol (SMP) File Management group."""

from __future__ import annotations

from enum import IntEnum, unique
from typing import Dict, Literal

from pydantic import BaseModel, ConfigDict

from smp import error, header, message


class FileDownloadRequest(message.ReadRequest):
    """Download contents of an existing file from specified path.

    Client applications must keep track of data they have already downloaded and
    where their position in the file is, and issue subsequent requests, with
    modified offset, to gather the entire file.

    Request does not carry size of requested chunk, the size is specified by
    application itself. Note that file handles will remain open for consecutive
    requests (as long as an idle timeout has not been reached and another
    transport does not make use of uploading/downloading files using fs_mgmt),
    but files are not exclusively owned by SMP, for the time of download
    session, and may change between requests or even be removed.
    """

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_DOWNLOAD_UPLOAD

    off: int
    """Offset in the file to read from."""
    name: str
    """Name of the file to download."""


class FileDownloadResponse(message.ReadResponse):
    """Success response to a file download request."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_DOWNLOAD_UPLOAD

    off: int
    """Offset that this response is for."""
    data: bytes
    """Data read from the file."""
    len: int | None = None
    """The length of the file, only mandatory if “off” is 0."""


class FileUploadRequest(message.WriteRequest):
    """Upload a file to a specified location.

    Command will automatically overwrite existing file or create a new one if it
    does not exist at specified path. The protocol supports stateless upload
    where each requests carries different chunk of a file and it is client side
    responsibility to track progress of upload.

    Note that file handles will remain open for consecutive requests (as long as
    an idle timeout has not been reached, but files are not exclusively owned by
    SMP, for the time of download session, and may change between requests or
    even be removed. Note that file handles will remain open for consecutive
    requests (as long as an idle timeout has not been reached and another
    transport does not make use of uploading/downloading files using fs_mgmt),
    but files are not exclusively owned by MCUmgr, for the time of download
    session, and may change between requests or even be removed.
    """

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_DOWNLOAD_UPLOAD

    off: int
    """Offset to start/continue writing to."""
    data: bytes
    """Data to write to the file."""
    name: str
    """Name of the file to upload."""
    len: int | None = None
    """The length of the file, only mandatory if “off” is 0."""


class FileUploadResponse(message.WriteResponse):
    """Success response to a file upload request."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_DOWNLOAD_UPLOAD

    off: int
    """Offset of the file."""


class FileStatusRequest(message.ReadRequest):
    """Retrieve status of an existing file from specified path of a target device."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_STATUS

    name: str


class FileStatusResponse(message.ReadResponse):
    """Success response to a file status request."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_STATUS

    len: int


class FileHashChecksumRequest(message.ReadRequest):
    """Generate a hash/checksum of an existing file at a specified path on a target device.

    Note that kernel heap memory is required for buffers to be allocated for
    this to function, and large stack memory buffers are required for generation
    of the output hash/checksum. Requires `CONFIG_MCUMGR_GRP_FS_CHECKSUM_HASH` to
    be enabled for the base functionality, supported hash/checksum are opt-in
    with `CONFIG_MCUMGR_GRP_FS_CHECKSUM_IEEE_CRC32` or
    `CONFIG_MCUMGR_GRP_FS_HASH_SHA256`.
    """

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_HASH_CHECKSUM

    name: str
    """Absolute path to the file to generate hash/checksum for."""
    type: Literal["crc32", "sha256"] | None = None
    """Type of hash/checksum to generate, if not provided, default is used."""
    off: int | None = None
    """Offset to start hash/checksum generation from, default 0."""
    len: int | None = None
    """Maximum length of the file to generate hash/checksum for, default is entire file."""


class FileHashChecksumResponse(message.ReadResponse):
    """Success response to a file hash/checksum request."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_HASH_CHECKSUM

    type: Literal["crc32", "sha256"]
    off: int | None = None
    """Only present if not 0."""
    len: int
    """Length of input data used to generate hash/checksum."""
    output: int | bytes
    """Output hash/checksum, depending on the type requested."""


class SupportedFileHashChecksumTypesRequest(message.ReadRequest):
    """List the hash and checksum types are available on a device.

    Requires Kconfig `CONFIG_MCUMGR_GRP_FS_CHECKSUM_HASH_SUPPORTED_CMD` to be enabled.
    """

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.SUPPORTED_FILE_HASH_CHECKSUM_TYPES


class HashChecksumFormat(IntEnum):
    """Format that the hash/checksum returns where 0 is for numerical and 1 is for byte array."""

    NUMERICAL = 0
    BYTE_ARRAY = 1


class HashChecksumType(BaseModel):
    """Hash and checksum type supported by the device."""

    model_config = ConfigDict(extra="forbid", frozen=True)

    format: HashChecksumFormat
    """Format that the hash/checksum returns where 0 is for numerical and 1 is for byte array."""
    size: int
    """Size of the hash/checksum in bytes."""


class SupportedFileHashChecksumTypesResponse(message.ReadResponse):
    """Success response to a supported file hash/checksum types request."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.SUPPORTED_FILE_HASH_CHECKSUM_TYPES

    types: Dict[Literal["crc32", "sha256"], HashChecksumType]
    """The map of supported hash/checksum types."""


class FileCloseRequest(message.WriteRequest):
    """Close all open file handles held by fs_mgmt upload/download requests."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_CLOSE


class FileCloseResponse(message.WriteResponse):
    """Success response to a file close request."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT
    _COMMAND_ID = header.CommandId.FileManagement.FILE_CLOSE


@unique
class FS_MGMT_ERR(IntEnum):
    """File System Management error codes."""

    OK = 0
    """No error (success)."""

    UNKNOWN = 1
    """ Unknown error. """

    FILE_INVALID_NAME = 2
    """The specified file name is not valid."""

    FILE_NOT_FOUND = 3
    """The specified file does not exist."""

    FILE_IS_DIRECTORY = 4
    """The specified file is a directory, not a file. """

    FILE_OPEN_FAILED = 5
    """Error occurred whilst attempting to open a file."""

    FILE_SEEK_FAILED = 6
    """Error occurred whilst attempting to seek to an offset in a file. """

    FILE_READ_FAILED = 7
    """Error occurred whilst attempting to read data from a file."""

    FILE_TRUNCATE_FAILED = 8
    """Error occurred whilst trying to truncate file."""

    FILE_DELETE_FAILED = 9
    """Error occurred whilst trying to delete file."""

    FILE_WRITE_FAILED = 10
    """Error occurred whilst attempting to write data to a file."""

    FILE_OFFSET_NOT_VALID = 11
    """The specified data offset is not valid, this could indicate that the file on the device has
    changed since the previous command. The length of the current file on the device is returned as
    "len", the user application needs to decide how to handle this (e.g. the hash of the file could
    be requested and compared with the hash of the length of the file being uploaded to see if they
    match or not)."""

    FILE_OFFSET_LARGER_THAN_FILE = 12
    """The requested offset is larger than the size of the file on the device."""

    CHECKSUM_HASH_NOT_FOUND = 13
    """The requested checksum or hash type was not found or is not supported by this build."""

    MOUNT_POINT_NOT_FOUND = 14
    """The specified mount point was not found or is not mounted."""

    READ_ONLY_FILESYSTEM = 15
    """The specified mount point is that of a read-only filesystem."""

    FILE_EMPTY = 16
    """The operation cannot be performed because the file is empty with no contents. """


class FileSystemManagementErrorV1(error.ErrorV1):
    """File System Management error response."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT


class FileSystemManagementErrorV2(error.ErrorV2[FS_MGMT_ERR]):
    """File System Management error response."""

    _GROUP_ID = header.GroupId.FILE_MANAGEMENT



"""The Simple Management Protocol (SMP) Image Management group.

## Notion of "slots" and "images" in Zephyr

The “slot” and “image” definition comes from mcuboot where “image” would consist
of two “slots”, further named “primary” and “secondary”; the application is
supposed to run from the “primary slot” and update is supposed to be uploaded to
the “secondary slot”; the mcuboot is responsible in swapping slots on boot. This
means that pair of slots is dedicated to single upgradable application. In case
of Zephyr this gets a little bit confusing because DTS will use
“slot0_partition” and “slot1_partition”, as label of fixed-partition dedicated
to single application, but will name them as “image-0” and “image-1”
respectively.

Currently Zephyr supports at most two images, in which case mapping is as
follows:

| Image | Slot labels | Slot Names |
|-------|-------------|------------|
| 0     | “slot0_partition” “slot1_partition” | “image-0” “image-1” |
| 1     | “slot2_partition” “slot3_partition” | “image-2” “image-3” |

"""

from __future__ import annotations

from enum import IntEnum, unique
from typing import Generator, List

from pydantic import BaseModel, ConfigDict, ValidationInfo, field_validator

from smp import error, header, message


class HashBytes(bytes):  # pragma: no cover
    """Only to print something useful to the `rich` console."""

    def __rich_repr__(self) -> Generator[str, None, None]:
        yield self.hex().upper()


class ImageState(BaseModel):
    """The state of an image in a slot."""

    model_config = ConfigDict(extra="forbid", frozen=True, arbitrary_types_allowed=True)

    slot: int
    """Slot number within “image”.

    Each image has two slots:
    - primary (running one) = 0
    - secondary (for DFU dual-bank purposes) = 1.
    """
    version: str
    """Version of the image."""
    image: int | None = None
    """Semi-optional image number.

    The field is not required when only one image is supported by the running
    application.
    """
    hash: HashBytes | bytes | None = None
    """SHA256 hash of the image header and body.

    Note that this will not be the same as the SHA256 of the whole file, it is
    the field in the MCUboot TLV section that contains a hash of the data which
    is used for signature verification purposes. This field is optional but only
    optional when using MCUboot's serial recovery feature with one pair of image
    slots, `CONFIG_BOOT_SERIAL_IMG_GRP_HASH` can be disabled to remove
    support for hashes in this configuration. SMP server in applications must
    support sending hashes.
    """
    bootable: bool | None = None
    """True if image has bootable flag set.

    This field does not have to be present if false.
    """
    pending: bool | None = None
    """True if image is set for next swap.

    This field does not have to be present if false.
    """
    confirmed: bool | None = None
    """True if image has been confirmed.

    This field does not have to be present if false.
    """
    active: bool | None = None
    """True if image is currently active application

    This field does not have to be present if false.
    """
    permanent: bool | None = None
    """True if image is to stay in primary slot after the next boot.

    This does not have to be present if false.
    """

    @field_validator("hash")
    @classmethod
    def cast_bytes(cls, v: bytes | None, _: ValidationInfo) -> HashBytes | None:
        if v is None:
            return None
        return HashBytes(v)


class ImageStatesReadRequest(message.ReadRequest):
    """Obtain list of images with their current state."""

    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT
    _COMMAND_ID = header.CommandId.ImageManagement.STATE


class ImageStatesReadResponse(message.ReadResponse):
    """Response to an image state request."""

    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT
    _COMMAND_ID = header.CommandId.ImageManagement.STATE

    images: List[ImageState]
    """List of images with their current state."""
    splitStatus: int | None = None
    """States whether loader of split image is compatible with application part.

    This is unused by Zephyr.
    """


class ImageStatesWriteRequest(message.WriteRequest):
    """Set the state of an image.

    If “confirm” is false or not provided, an image with the “hash” will be set
    for test, which means that it will not be marked as permanent and upon hard
    reset the previous application will be restored to the primary slot. In case
    when “confirm” is true, the “hash” is optional as the currently running
    application will be assumed as target for confirmation.
    """

    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT
    _COMMAND_ID = header.CommandId.ImageManagement.STATE

    hash: bytes | None = None
    """SHA256 hash of the image header and body."""
    confirm: bool = False
    """Confirm the image given by hash.

    CAUTION: it is dangerous to confirm the image before it has been booted!

    Setting this to true will mark the image as confirmed before it has been
    booted, which can brick the device.  Zephyr provides hooks for confirming
    the image from within the application, so this is not necessary for normal
    operation.
    """


class ImageStatesWriteResponse(ImageStatesReadResponse):
    """Success response to an image state write request."""


class ImageUploadWriteRequest(message.WriteRequest):
    """Upload an image to the device.

    The image is uploaded in chunks, with each chunk being sent in a separate
    request. The first request must include the image's length and image number.
    """

    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT
    _COMMAND_ID = header.CommandId.ImageManagement.UPLOAD

    off: int
    """The offset of the data chunk in this request."""
    data: bytes
    """The data chunk to write."""
    image: int | None = None
    """Image number.

    It does not have to appear in request at all, in which case it is assumed to
    be 0. Should only be present when `off` is 0.
    """
    len: int | None = None
    """The length of the image.

    Required when `off` is 0, otherwise ignored.
    """
    sha: bytes | None = None
    """SHA256 hash of the image.

    This is used to identify an upload session (e.g. to allow to continue
    a broken session), and for image verification purposes. This must be a full
    SHA256 hash of the whole image being uploaded, or not included if the hash
    is not available (in which case, upload session continuation and image
    verification functionality will be unavailable). Should only be present when
    “off” is 0.
    """
    upgrade: bool | None = None
    """Optional flag that states that only upgrade should be allowed.

    If the version of uploaded software is not higher then already on a device,
    the image upload will be rejected. Zephyr compares major, minor and
    revision (x.y.z) by default unless `CONFIG_MCUMGR_GRP_IMG_VERSION_CMP_USE_BUILD_NUMBER`
    is set, whereby it will compare build numbers too. Should only be present
    when “off” is 0.
    """


class ImageUploadWriteResponse(message.WriteResponse):
    """Success response to an image upload request."""

    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT
    _COMMAND_ID = header.CommandId.ImageManagement.UPLOAD

    off: int | None = None
    """The portion of the upload that has been completed, in 8-bit bytes.

    This is the offset of the next byte to be written. If the offset is equal to
    the length of the image, the upload is complete.
    """

    match: bool | None = None
    """Indicates if the uploaded data successfully matches the provided SHA256.

    Only sent in the final packet if CONFIG_IMG_ENABLE_IMAGE_CHECK is enabled.
    """

    rc: int | None = None
    """Legacy field that contains a return code; possibly `MGMT_ERR`.

    This field may be present on old SMP server implementations or new SMP
    server implementations that have set
    `CONFIG_MCUMGR_SMP_LEGACY_RC_BEHAVIOUR=y` for backwards compatibility with
    old SMP clients.

    Note that we are not validating this field because we don't necessarily
    trust the server to send us valid values. If this value is present, then it
    indicates use of an SMP server that is out of spec and interpretation of the
    value should be done with reference to that server's source code, rather
    that the SMP specification.

    Zephyr source code reference: https://github.com/zephyrproject-rtos/zephyr/blob/91a1e706535b2f99433280513c5bc66dfb918506/subsys/mgmt/mcumgr/grp/img_mgmt/src/img_mgmt.c#L397-L400
    """  # noqa: E501


class ImageEraseRequest(message.WriteRequest):
    """Erase an image from a slot."""

    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT
    _COMMAND_ID = header.CommandId.ImageManagement.ERASE

    slot: int | None = None
    """The slot to erase. If not provided, slot 1 will be erased."""


class ImageEraseResponse(message.WriteResponse):
    """Success response to an image erase request."""

    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT
    _COMMAND_ID = header.CommandId.ImageManagement.ERASE


@unique
class IMG_MGMT_ERR(IntEnum):
    """Image management error codes."""

    OK = 0
    """No error, this is implied if there is no ret value in the response"""

    UNKNOWN = 1
    """Unknown error occurred."""

    FLASH_CONFIG_QUERY_FAIL = 2
    """Failed to query flash area configuration."""

    NO_IMAGE = 3
    """There is no image in the slot."""

    NO_TLVS = 4
    """The image in the slot has no TLVs (tag, length, value)."""

    INVALID_TLV = 5
    """The image in the slot has an invalid TLV type and/or length."""

    TLV_MULTIPLE_HASHES_FOUND = 6
    """The image in the slot has multiple hash TLVs, which is invalid."""

    TLV_INVALID_SIZE = 7
    """The image in the slot has an invalid TLV size."""

    HASH_NOT_FOUND = 8
    """The image in the slot does not have a hash TLV, which is required."""

    NO_FREE_SLOT = 9
    """There is no free slot to place the image."""

    FLASH_OPEN_FAILED = 10
    """Flash area opening failed."""

    FLASH_READ_FAILED = 11
    """Flash area reading failed."""

    FLASH_WRITE_FAILED = 12
    """Flash area writing failed."""

    FLASH_ERASE_FAILED = 13
    """Flash area erase failed."""

    INVALID_SLOT = 14
    """The provided slot is not valid."""

    NO_FREE_MEMORY = 15
    """Insufficient heap memory (malloc failed)."""

    FLASH_CONTEXT_ALREADY_SET = 16
    """The flash context is already set."""

    FLASH_CONTEXT_NOT_SET = 17
    """The flash context is not set."""

    FLASH_AREA_DEVICE_NULL = 18
    """The device for the flash area is NULL."""

    INVALID_PAGE_OFFSET = 19
    """The offset for a page number is invalid."""

    INVALID_OFFSET = 20
    """The offset parameter was not provided and is required."""

    INVALID_LENGTH = 21
    """The length parameter was not provided and is required."""

    INVALID_IMAGE_HEADER = 22
    """The image length is smaller than the size of an image header."""

    INVALID_IMAGE_HEADER_MAGIC = 23
    """The image header magic value does not match the expected value."""

    INVALID_HASH = 24
    """The hash parameter provided is not valid."""

    INVALID_FLASH_ADDRESS = 25
    """The image load address does not match the address of the flash area."""

    VERSION_GET_FAILED = 26
    """Failed to get version of currently running application."""

    CURRENT_VERSION_IS_NEWER = 27
    """The currently running application is newer than the version being uploaded."""

    IMAGE_ALREADY_PENDING = 28
    """There is already an image operating pending."""

    INVALID_IMAGE_VECTOR_TABLE = 29
    """The image vector table is invalid."""

    INVALID_IMAGE_TOO_LARGE = 30
    """The image it too large to fit."""

    INVALID_IMAGE_DATA_OVERRUN = 31
    """The amount of data sent is larger than the provided image size."""

    IMAGE_CONFIRMATION_DENIED = 32
    """Confirmation of image has been denied"""

    IMAGE_SETTING_TEST_TO_ACTIVE_DENIED = 33
    """Setting test to active slot is not allowed"""


class ImageManagementErrorV1(error.ErrorV1):
    """Image Management error response."""

    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT


class ImageManagementErrorV2(error.ErrorV2[IMG_MGMT_ERR]):
    """Image Management error response."""

    _GROUP_ID = header.GroupId.IMAGE_MANAGEMENT



"""Simple Management Protocol (SMP) runtime Python `Exception`s."""


class SMPException(Exception):
    """Base class for SMP exceptions."""


class SMPDecodeError(SMPException):
    """An error occurred while decoding an SMP message."""


class SMPBadStartDelimiter(SMPDecodeError):
    """The start delimiter of the SMP message is invalid."""


class SMPBadContinueDelimiter(SMPDecodeError):
    """The continue delimiter of the SMP message is invalid."""


class SMPBadCRC(SMPDecodeError):
    """The CRC of the SMP message failed."""


class SMPDeserializationError(SMPException):
    """Failed to deserialize an SMP message."""


class SMPMismatchedGroupId(SMPDeserializationError):
    """Group ID in the SMP message does not match the expected group."""


class SMPMalformed(SMPException):
    """The SMP message is malformed."""



"""Generators for encoding and decoding Simple Management Protocol (SMP) packets."""

import logging
import struct
from base64 import b64decode, b64encode
from typing import Final, Generator

from crcmod.predefined import mkPredefinedCrcFun  # type: ignore

from smp.exceptions import SMPBadContinueDelimiter, SMPBadCRC, SMPBadStartDelimiter

SIXTY_NINE: Final = struct.pack("!BB", 6, 9)
START_DELIMITER: Final = SIXTY_NINE
FOUR_TWENTY: Final = struct.pack("!BB", 4, 20)
CONTINUE_DELIMITER: Final = FOUR_TWENTY
DELIMITER_SIZE: Final = len(SIXTY_NINE)
assert len(SIXTY_NINE) == len(FOUR_TWENTY) and len(FOUR_TWENTY) == DELIMITER_SIZE
LF: Final = b"\n"
END_DELIMITER: Final = LF
LINE_LENGTH_SUBTRACTOR: Final = 4  # you'd think it's 3, len(delimiter) + len(CR)
FRAME_LENGTH_STRUCT: Final = struct.Struct("!H")
CRC16_STRUCT: Final = struct.Struct("!H")

crc16_func: Final = mkPredefinedCrcFun("xmodem")


logger: Final = logging.getLogger(__name__)


def encode(request: bytes, line_length: int = 127) -> Generator[bytes, None, None]:
    """Iteratively pack an SMP bytes to packets of `line_length` size.

    Note: only the USB/serial transport uses this encoding and fragmentation.
    The BLE and UDP transports can simply send the SMP frames as they are.

    Params:
    - request: The bytes to be encoded.
    - line_length: The maximum length of each packet.

    Example:

    Encode arbitrary bytes to SMP packets:

    ```python
    from smp.packet import encode

    for packet in encode(b"Hello, world!"):
        print(packet)
    ```
    prints:
    ```
    b'\\x06\\tAA9IZWxsbywgd29ybGQhet4=\\n'
    ```

    Set a smaller packet size to fragment:

    ```python
    from smp.packet import encode

    for packet in encode(b"Hello, world!", line_length=8):
        print(packet)
    ```
    prints:
    ```
    b'\\x06\\tAA9I\\n'
    b'\\x04\\x14ZWxs\\n'
    b'\\x04\\x14bywg\\n'
    b'\\x04\\x14d29y\\n'
    b'\\x04\\x14bGQh\\n'
    b'\\x04\\x14et4=\\n'
    ```

    The packet size should be set according to the buffer sizes of the SMP
    server, and larger buffers will be more efficient.  It is not necessary to
    use multiple buffers.
    """

    logger.debug(f"Serializing {request=}")

    crc16 = CRC16_STRUCT.pack(crc16_func(request))
    frame_length = FRAME_LENGTH_STRUCT.pack(len(request) + CRC16_STRUCT.size)

    total_size = FRAME_LENGTH_STRUCT.size + len(request) + CRC16_STRUCT.size
    logger.debug(f"Total size of request is {total_size}B")

    packet_size = ((line_length - LINE_LENGTH_SUBTRACTOR) // 4) * 4

    complete_b64 = b64encode(frame_length + request + crc16)
    logger.debug(f"Total size of b64 encoded request is {len(complete_b64)}B")

    # send the start delimiter, as many bytes as possible, and newline
    yield START_DELIMITER + complete_b64[:packet_size] + END_DELIMITER

    remaining = complete_b64[packet_size:]

    while len(remaining) > 0:
        # send the continue delimiter, as many bytes as possible, and newline
        yield CONTINUE_DELIMITER + remaining[:packet_size] + END_DELIMITER
        remaining = remaining[packet_size:]


def decode() -> Generator[None, bytes, bytes]:
    """Iteratively unpack a series of SMP packets to SMP frame bytes.

    Example:

    ```python
    from smp.packet import decode
    from smp.image_management import ImageUploadWriteRequest

    # Example encoded SMP packets
    packets = (
        bytes.fromhex("06094148344b41414230414147610a"),
        bytes.fromhex("041441615a6a6247567544574e760a"),
        bytes.fromhex("04145a6d59594b6d4e7a6147464d0a"),
        bytes.fromhex("041464484a3163335167625755670a"),
        bytes.fromhex("0414596e4a765a475268644746590a"),
        bytes.fromhex("04145141414241674d45425159480a"),
        bytes.fromhex("041443416b4b4377774e446738510a"),
        bytes.fromhex("041445524954464255574678675a0a"),
        bytes.fromhex("04144768736348523466494345690a"),
        bytes.fromhex("04144979516c4a69636f4b536f720a"),
        bytes.fromhex("04144c4330754c7a41784d6a4d300a"),
        bytes.fromhex("04144e5459334f446b364f7a77390a"),
        bytes.fromhex("0414506a396c615731685a3255410a"),
        bytes.fromhex("04145a3356775a334a685a4758300a"),
        bytes.fromhex("04147134673d0a"),
    )

    decoder = decode()
    next(decoder)

    for packet in packets:
        try:
            decoder.send(packet)
        except StopIteration as e:
            frame = e.value
            break

    print(frame)
    print(ImageUploadWriteRequest.loads(frame))
    ```
    prints:
    ```
    bytearray(b'\\n\\x00\\x00t\\x00\\x01\\x9a\\x01\\xa6clen\\rcoff\\x18*cshaLtrust me broddataX@\\x00\\x01\\x02\\x03\\x04\\x05\\x06\\x07\\x08\\t\\n\\x0b\\x0c\\r\\x0e\\x0f\\x10\\x11\\x12\\x13\\x14\\x15\\x16\\x17\\x18\\x19\\x1a\\x1b\\x1c\\x1d\\x1e\\x1f !"#$%&\\'()*+,-./0123456789:;<=>?eimage\\x00gupgrade\\xf4')  # noqa: E501
    ```
    and
    ```
    header=Header(op=<OP.WRITE: 2>, version=<Version.V2: 1>, flags=<Flag: 0>,
    length=116, group_id=1, sequence=154, command_id=1) version=<Version.V2: 1>
    sequence=154 smp_data=b'\\n\\x00\\x00t\\x00\\x01\\x9a\\x01\\xa6clen\\rcoff\\x18*cshaL
    trust me broddataX@\\x00\\x01\\x02\\x03\\x04\\x05\\x06\\x07\\x08\\t\\n\\x0b\\x0c\\r\\x0e
    \\x0f\\x10\\x11\\x12\\x13\\x14\\x15\\x16\\x17\\x18\\x19\\x1a\\x1b\\x1c\\x1d\\x1e\\x1f !"#$%&
    \\'()*+,-./0123456789:;<=>?eimage\\x00gupgrade\\xf4' off=42 data=b'\\x00\\x01\\x02
    \\x03\\x04\\x05\\x06\\x07\\x08\\t\\n\\x0b\\x0c\\r\\x0e\\x0f\\x10\\x11\\x12\\x13\\x14\\x15\\x16
    \\x17\\x18\\x19\\x1a\\x1b\\x1c\\x1d\\x1e\\x1f !"#$%&\\'()*+,-./0123456789:;<=>?'
    image=0 len=13 sha=b'trust me bro' upgrade=False
    ```
    """

    packet = yield

    if packet[:DELIMITER_SIZE] != START_DELIMITER:
        raise SMPBadStartDelimiter(f"Bad start delimiter {packet[:DELIMITER_SIZE].hex()}")

    length_and_frame = b64decode(packet[DELIMITER_SIZE : -len(END_DELIMITER)])
    frame_length = FRAME_LENGTH_STRUCT.unpack(length_and_frame[: FRAME_LENGTH_STRUCT.size])[0]
    logger.debug(f"Response {length_and_frame=}")

    frame = bytearray(length_and_frame[FRAME_LENGTH_STRUCT.size :])
    logger.debug(f"length of frame: {len(frame)}")
    while len(frame) < frame_length:
        packet = yield

        if packet[:DELIMITER_SIZE] != CONTINUE_DELIMITER:
            raise SMPBadContinueDelimiter(f"Bad continue delimiter {packet[:DELIMITER_SIZE].hex()}")

        frame.extend(b64decode(packet[DELIMITER_SIZE : -len(END_DELIMITER)]))

    complete_frame = frame[: -CRC16_STRUCT.size]
    packet_crc16 = CRC16_STRUCT.unpack(frame[-CRC16_STRUCT.size :])[0]
    calculated_crc16 = crc16_func(complete_frame)

    if packet_crc16 != calculated_crc16:
        raise SMPBadCRC(f"Packet CRC {hex(packet_crc16)} != calculated CRC {hex(calculated_crc16)}")

    return complete_frame
